/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <dlog.h>
#include <wayland-client.h>
#include <Ecore_Wayland.h>
#include <Elementary.h>
#include <wayland-extension/tizen-extension-client-protocol.h>
#include <tizen-remote-surface-client-protocol.h>
#include <aul_screen_connector.h>

#include "screen_connector_trs.h"
#include "screen_connector_watcher.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_WATCHER"

struct _screen_connector_watcher_h {
	void (*update_cb)(struct tizen_remote_surface *trs, struct wl_buffer *buffer, uint32_t time, void *data);
	void (*missing_cb)(struct tizen_remote_surface *trs, void *data);
	void (*change_cb)(struct tizen_remote_surface *trs, uint32_t type, struct wl_buffer *tbm,
			int32_t img_file_fd, uint32_t img_file_size, uint32_t time, struct wl_array *keys, void *data);
	screen_connector_watcher_ops *ops;
	void *data;
	GHashTable *info_table;
	aul_screen_viewer_h aul_handle;
};

struct _screen_connector_info_h {
	char *appid;
	char *instance_id;
	int pid;
	int surface_id;
	screen_connector_watcher_ops *ops;
	struct tizen_remote_surface *surface;
	bool renderable;
	void *data;
};

static void __buffer_updated(void *data, struct tizen_remote_surface *trs, struct wl_buffer *buffer, uint32_t time)
{
}

static void __buffer_missing(void *data, struct tizen_remote_surface *trs)
{
	screen_connector_info_h info = (screen_connector_info_h)data;

	if (info)
		info->ops->removed_cb(info->appid, info->instance_id, info->pid, info->data);
}

static void __buffer_changed(void *data, struct tizen_remote_surface *trs, uint32_t type, struct wl_buffer *tbm,
		int32_t img_file_fd, uint32_t img_file_size, uint32_t time, struct wl_array *keys)
{
	screen_connector_info_h info = (screen_connector_info_h)data;

	if (info && info->ops) {

		if (info->renderable) {
			info->renderable = false;
			tizen_remote_surface_set_remote_render(info->surface, 0);
		}

		info->ops->updated_cb(trs, type, tbm, img_file_fd, img_file_size, time, keys, info->appid,
				info->instance_id, info->pid, info->data);
	}
}

static const struct tizen_remote_surface_listener __rs_listener = {
	__buffer_updated,
	__buffer_missing,
	__buffer_changed,
};

EXPORT_API int screen_connector_watcher_redirect_surface(screen_connector_info_h info)
{

	struct wl_tbm *tbm = screen_connector_get_tbm();
	struct tizen_remote_surface_manager *rsm = screen_connector_get_rsm();
	if (info == NULL || rsm == NULL || tbm == NULL)
		return -1;

	info->surface = tizen_remote_surface_manager_create_surface(rsm, (uint32_t)info->surface_id, tbm);
	if (info->surface == NULL)
		return -1;

	LOGD("resource_id : %d, (%p)", info->surface_id, info->surface);
	tizen_remote_surface_add_listener(info->surface, &__rs_listener, info);
	tizen_remote_surface_redirect(info->surface);

	return 0;
}

static void __destroy_screen_info(gpointer data)
{
	screen_connector_info_h info = (screen_connector_info_h)data;

	if (info == NULL)
		return;

	LOGD("__destroy_screen_info appid: %s surface_id %d", info->appid, info->surface_id);

	if (info->appid) {
		free(info->appid);
		info->appid = NULL;
	}

	if (info->instance_id) {
		free(info->instance_id);
		info->instance_id = NULL;
	}

	if (info->surface) {
		if (info->renderable)
			tizen_remote_surface_set_remote_render(info->surface, 0);

		tizen_remote_surface_set_user_data(info->surface, NULL);
		tizen_remote_surface_unredirect(info->surface);
		tizen_remote_surface_destroy(info->surface);
		info->surface = NULL;
	}

	free(info);
}

static void __aul_screen_viewer_cb(const char *appid, const char *instance_id,
		const int pid, const unsigned int surface_id,
		aul_screen_connector_event_type_e event_type, void *data)
{
	screen_connector_watcher_h watcher_h = (screen_connector_watcher_h)data;
	screen_connector_info_h info = (screen_connector_info_h)calloc(1, sizeof(struct _screen_connector_info_h));
	screen_connector_info_h cur_info;

	LOGD("__aul_screen_viewer_cb pid: %d, surface_id: %d, appid: %s type %d", pid, surface_id, appid, event_type);
	info->appid = strdup(appid);
	info->instance_id = strdup(instance_id);
	info->pid = pid;
	info->surface_id = surface_id;
	info->ops = watcher_h->ops;
	info->data = watcher_h->data;
	info->renderable = false;

	if (event_type == AUL_SCREEN_CONNECTOR_EVENT_TYPE_ADD) {

		g_hash_table_insert(watcher_h->info_table,
				strdup(instance_id), info);

		info->ops->added_cb(appid, instance_id, pid, info->data);
		screen_connector_watcher_redirect_surface(info);

	} else if (event_type == AUL_SCREEN_CONNECTOR_EVENT_TYPE_UPDATE) {

		cur_info = g_hash_table_lookup(watcher_h->info_table, instance_id);
		if (cur_info && cur_info->surface_id != surface_id) {
			cur_info->ops->removed_cb(appid, instance_id, pid, cur_info->data);
			g_hash_table_remove(watcher_h->info_table, instance_id);
		}

		g_hash_table_insert(watcher_h->info_table, strdup(instance_id), info);

		info->ops->added_cb(appid, instance_id, pid, info->data);
		screen_connector_watcher_redirect_surface(info);
	} else if (event_type == AUL_SCREEN_CONNECTOR_EVENT_TYPE_REMOVE) {

		info->ops->removed_cb(appid, instance_id, pid, info->data);
		g_hash_table_remove(watcher_h->info_table, instance_id);
		__destroy_screen_info(info);
	}
}

EXPORT_API int screen_connector_watcher_init(void)
{
	int ret = 0;
	ret = screen_connector_trs_init();
	return ret;
}

EXPORT_API int screen_connector_watcher_fini(void)
{
	int ret = 0;
	ret = screen_connector_trs_fini();
	return ret;
}

static void __destroy_watcher_h(screen_connector_watcher_h handle)
{
	if (!handle)
		return;

	if (handle->ops)
		free(handle->ops);

	if (handle->aul_handle)
		aul_screen_connector_remove_screen_viewer(handle->aul_handle);

	if (handle->info_table)
		g_hash_table_destroy(handle->info_table);

	free(handle);
}

EXPORT_API screen_connector_watcher_h screen_connector_watcher_add(screen_connector_watcher_ops *ops, void *data)
{
	screen_connector_watcher_h watcher_h = NULL;
	aul_screen_viewer_h handle = NULL;

	watcher_h = (screen_connector_watcher_h)calloc(1, sizeof(struct _screen_connector_watcher_h));
	if (watcher_h == NULL) {
		LOGE("watcher_h calloc fail");
		return NULL;
	}
	watcher_h->data = data;
	watcher_h->ops = ops;
	watcher_h->info_table = g_hash_table_new_full(g_str_hash,
			g_str_equal, free, __destroy_screen_info);
	aul_screen_connector_add_screen_viewer(__aul_screen_viewer_cb, AUL_SCREEN_TYPE_UI, false, watcher_h, &handle);
	watcher_h->aul_handle = handle;

	return watcher_h;
}

EXPORT_API int screen_connector_watcher_remove(screen_connector_watcher_h h)
{
	if (h == NULL) {
		LOGE("screen_connector_watcher_h is NULL");
		return -1;
	}
	__destroy_watcher_h(h);

	return 0;
}

EXPORT_API int screen_connector_watcher_update(screen_connector_watcher_h h, const char *instance_id)
{
	screen_connector_info_h info;

	if (h == NULL) {
		LOGE("toolkit_h is NULL");
		return -1;
	}

	if (h->info_table == NULL) {
		LOGE("info_table is NULL");
		return -1;
	}

	info = (screen_connector_info_h)g_hash_table_lookup(h->info_table, instance_id);
	if (info == NULL) {
		LOGE("invalid instance_id %s", instance_id);
		return -1;
	}

	LOGI("send update request to %s", instance_id);
	tizen_remote_surface_set_remote_render(info->surface, 1);
	/* flag will be cleared when buffer changed event received */
	info->renderable = true;

	return aul_screen_connector_send_update_request(info->appid, info->instance_id);
}

EXPORT_API char *screen_connector_watcher_get_appid(screen_connector_info_h h)
{
	return h->appid;
}

EXPORT_API char *screen_connector_watcher_get_instance_id(screen_connector_info_h h)
{
	return h->instance_id;
}

EXPORT_API int screen_connector_watcher_get_pid(screen_connector_info_h h)
{
	return h->pid;
}

EXPORT_API int screen_connector_watcher_get_surface_id(screen_connector_info_h h)
{
	return h->surface_id;
}

EXPORT_API void *screen_connector_watcher_get_user_data(screen_connector_info_h h)
{
	return h->data;
}
